package edu.ncsu.AdhocClient;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.StringTokenizer;

import android.app.AlertDialog;
import android.app.Application;
import android.content.Context;
import android.content.DialogInterface;
import android.content.SharedPreferences;
import android.location.Location;
import android.location.LocationManager;
import android.net.wifi.WifiInfo;
import android.net.wifi.WifiManager;
import android.os.Looper;
import android.os.Message;
import android.preference.PreferenceManager;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.widget.Toast;

public class AdhocClient extends Application {
    private static final String MSG_TAG = "AdhocClient";
	public static final long SERVER_TIME = 60000; // 60 seconds
	public static final long CLIENT_TIME = 60000; // 60 seconds

	private Thread netControlThread = null;
	private Thread clientSendThread = null;
	private Thread clientReceiveThread = null;

	// WifiManager
	private WifiManager wifiManager;

	private LocationManager locationManager;
	
	// CoreTask
	public CoreTask coretask = null;

	// Preferences
	public SharedPreferences settings = null;
	public SharedPreferences.Editor preferenceEditor = null;

	private List<String> eventList = new ArrayList<String>();
	private List<String> friendList = new ArrayList<String>();
	private List<String> messageHeader = new ArrayList<String>();
	private List<String> messageList = new ArrayList<String>();
	public int curMessageIndex = -1;
	public int curFriendIndex = -1;
	
	class FriendData {
		public long lastHeartbeat=0;
		public String name;
		public String IPaddress;
		public String statusMsg;
	
		FriendData() { }
		FriendData(String _name, String _IPaddress, String _statusMsg) {
			name=_name;
			IPaddress=_IPaddress;
			statusMsg=_statusMsg;
		}
		boolean equals(FriendData other) {
			return name.compareTo(other.name) == 0 && 
				   IPaddress.compareTo(other.IPaddress) == 0;
		}
		String makeHeartbeat(String gpsStr) {
			String hbString=new String();
			hbString  = "Heartbeat:~" + name + "~" +
			            statusMsg + "~" + gpsStr;
			return hbString;
		}
		boolean isHeartbeat(String msg) {
			if (msg.startsWith("Heartbeat:"))
				return true;
			else
				return false;
		}
		String parseHeartbeat(String hbString) {
			StringTokenizer st = new StringTokenizer(hbString, "~");
		    if (st.hasMoreTokens()) {
		    	// Skip message identifier
		    	st.nextToken();
		    }
			if (st.hasMoreTokens()) {
		         name=st.nextToken();
		    }
		    if (st.hasMoreTokens()) {
		         statusMsg=st.nextToken();
		    }
		    String gps="";
		    if (st.hasMoreTokens()) {
		    	gps=st.nextToken();
		    }
		    return gps;
		}
		String makeMessage(String message) {
			String msgString=new String();
			msgString  = "Message:~" + name + "~";
			msgString += message + "~";
			return msgString;
		}
		boolean isMessage(String msg) {
			if (msg.startsWith("Message:"))
				return true;
			else
				return false;
		}
		String parseMessage(String msgString) {
			StringTokenizer st = new StringTokenizer(msgString, "~");
		    if (st.hasMoreTokens()) {
		    	// Skip message identifier
		    	st.nextToken();
		    }
			if (st.hasMoreTokens()) {
		         name=st.nextToken();
		    }
		    String msgText="";
		    if (st.hasMoreTokens()) {
		         msgText=st.nextToken();
		    }
		    return msgText;
		}
	}
	
	public FriendData myData;
	
	private HashMap<Integer, FriendData> activeFriends = new HashMap<Integer, FriendData>();

	// Network Control state
	public enum netControlState { NOT_RUNNING, ADHOC_SERVER, ADHOC_CLIENT, ACCESS_POINT };
	private netControlState curState;
	long lastSwitchTime=0;
	boolean discovered=false;
	boolean serviceStarted=false;
	int staleFriendCount=0;
	String myDhcpAddress;
	double dummyLat=0;
	double dummyLong=0;
	
    /** Called when the activity is first created. */
    public void onCreate() {
	    
		coretask = new CoreTask();
		coretask.setPath(this.getApplicationContext().getFilesDir().getParent());
		Log.d(MSG_TAG, "Current directory is "+this.getApplicationContext().getFilesDir().getParent());

		// Preferences
		settings = PreferenceManager.getDefaultSharedPreferences(this);
		
        // preferenceEditor
        this.preferenceEditor = settings.edit();

        // init wifiManager
        wifiManager = (WifiManager) this.getSystemService(Context.WIFI_SERVICE); 
		
        // init locationManager
        locationManager = (LocationManager)getSystemService(Context.LOCATION_SERVICE);
        
        // Check Homedir, or create it
        this.checkDirs(); 

        // Check for binaries
        boolean filesetOutdated = coretask.filesetOutdated();
        if (binariesExists() == false || filesetOutdated) {
        	if (coretask.hasRootPermission()) {
        		installBinaries();
        		//if (filesetOutdated) {
        			//this.openConfigRecoverDialog();
        		//}
        	}
        	else {
        		this.openNotRootDialog();
        	}
        }
    
        curState = netControlState.NOT_RUNNING;
        
        this.myData=new FriendData();
        
        this.eventList = new ArrayList<String>();
        this.friendList = new ArrayList<String>();
       
	    this.netControlThread = new Thread(new NetworkControl());
        this.netControlThread.start(); 
        
	    this.clientSendThread = new Thread(new ClientSender());
        this.clientSendThread.start(); 
        
	    this.clientReceiveThread = new Thread(new ClientReceiver());
        this.clientReceiveThread.start(); 

        setMyStatusMsg("Available");
        startService();
    }
    
    public void onTerminate() {
    	this.netControlThread.interrupt();
    	this.clientSendThread.interrupt();
    	this.clientReceiveThread.interrupt();
	}
    
    // From TetherApplication.java
    
    // Binary install
    public boolean binariesExists() {
    	File file = new File(this.coretask.DATA_FILE_PATH+"/bin/tether");
    	if (file.exists()) {
    		return true;
    	}
    	return false;
    }
    
    public void installBinaries() {
    	List<String> filenames = new ArrayList<String>();
    	// tether
    	this.copyBinary(this.coretask.DATA_FILE_PATH+"/bin/netcontrol", R.raw.netcontrol);
    	filenames.add("netcontrol");
    	// dnsmasq
    	this.copyBinary(this.coretask.DATA_FILE_PATH+"/bin/dnsmasq", R.raw.dnsmasq);
    	filenames.add("dnsmasq");
    	try {
			this.coretask.chmodBin(filenames);
		} catch (Exception e) {
			displayToastMessage("Unable to change permission on binary files!");
		}
    	// dnsmasq.conf
    	this.copyBinary(this.coretask.DATA_FILE_PATH+"/conf/dnsmasq.conf", R.raw.dnsmasq_conf);
    	// tiwlan.ini
    	this.copyBinary(this.coretask.DATA_FILE_PATH+"/conf/tiwlan.ini", R.raw.tiwlan_ini);
    	this.displayToastMessage("Binaries and config-files installed!");
    }
    
    private void copyBinary(String filename, int resource) {
    	File outFile = new File(filename);
    	InputStream is = this.getResources().openRawResource(resource);
    	byte buf[]=new byte[1024];
        int len;
        try {
        	OutputStream out = new FileOutputStream(outFile);
        	while((len = is.read(buf))>0) {
				out.write(buf,0,len);
			}
        	out.close();
        	is.close();
		} catch (IOException e) {
			this.displayToastMessage("Couldn't install file - "+filename+"!");
		}
    }
    

    private void checkDirs() {
    	File dir = new File(this.coretask.DATA_FILE_PATH);
    	if (dir.exists() == false) {
    			this.displayToastMessage("Application data-dir does not exist!");
    	}
    	else {
	    	dir = new File(this.coretask.DATA_FILE_PATH+"/bin");
	    	if (dir.exists() == false) {
	    		if (!dir.mkdir()) {
	    			this.displayToastMessage("Couldn't create bin-directory!");
	    		}
	    	}
	    	dir = new File(this.coretask.DATA_FILE_PATH+"/var");
	    	if (dir.exists() == false) {
	    		if (!dir.mkdir()) {
	    			this.displayToastMessage("Couldn't create var-directory!");
	    		}
	    	}
	    	dir = new File(this.coretask.DATA_FILE_PATH+"/conf");
	    	if (dir.exists() == false) {
	    		if (!dir.mkdir()) {
	    			this.displayToastMessage("Couldn't create conf-directory!");
	    		}
	    	}   			
    	}
    }    

    public static String now() {
        Calendar cal = Calendar.getInstance();
        SimpleDateFormat sdf = new SimpleDateFormat("H:mm:ss");
        return sdf.format(cal.getTime());
    }
    
    private void notifyIncomingEventActivity(){
    	if (IncomingEventActivity.currentInstance != null){
    		IncomingEventActivity.currentInstance.updateHandler.sendMessage(new Message());
    	}
    }

    private void notifyFriendListActivity(){
    	if (FriendListActivity.currentInstance != null){
    		FriendListActivity.currentInstance.updateHandler.sendMessage(new Message());
    	}
    }
    
    private void notifyStatusActivity(){
    	if (MyStatusActivity.currentInstance != null){
    		MyStatusActivity.currentInstance.updateHandler.sendMessage(new Message());
    	}
    }
    public synchronized void addEvent(String eventText) {
    	eventList.add(0, now()+" "+eventText);
    	notifyIncomingEventActivity();
    }
    
    // Display Toast-Message
	public void displayToastMessage(String message) {
		Toast.makeText(this, message, Toast.LENGTH_LONG).show();
	}
    
	// From MainActivity.java
	private void openNotRootDialog() {
		LayoutInflater li = LayoutInflater.from(this);
        View view = li.inflate(R.layout.norootview, null); 
		new AlertDialog.Builder(this)
        .setTitle("Not Root!")
        .setIcon(R.drawable.warning)
        .setView(view)
        .setNegativeButton("Close", new DialogInterface.OnClickListener() {
                public void onClick(DialogInterface dialog, int whichButton) {
                        Log.d(MSG_TAG, "Close pressed");
                        //this.application.finish();
                }
        })
        .setNeutralButton("Override", new DialogInterface.OnClickListener() {
                public void onClick(DialogInterface dialog, int whichButton) {
                    Log.d(MSG_TAG, "Override pressed");
                    installBinaries();
                }
        })
        .show();
   	}
	
    // get preferences
    public boolean getAdhocMode() {
    	int mode=Integer.parseInt(this.settings.getString("wifimode", "0"));
    	return (mode == 0 ? true : false);
    }
    
    public String getUserName() {
    	String n=this.settings.getString("username", "user");
    	if (n.length()==0)
    		n="user";
    	return n;
    }
	
    public void setMyIPAddress(String IPaddress) {
    	myData.name=getUserName();
    	myData.IPaddress=IPaddress;	    	
    }
    
    public void setMyStatusMsg(String statusMessage) {
    	myData.name=getUserName();
    	myData.statusMsg=statusMessage;
    }
    
    public String getMyStatusMsg() {
    	return myData.statusMsg;
    }
    
    // Wifi
    public void disableWifi(netControlState nextState) {
    	boolean done=false;
    	while (!done && !Thread.currentThread().isInterrupted()) {
			this.wifiManager.setWifiEnabled(false);
			// Waiting for interface-shutdown
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				// nothing
			}
			if (this.wifiManager.isWifiEnabled())
				done=false;
			else
				done=true;
    	}
		Log.d(MSG_TAG, "Wifi disabled!");
		curState=nextState;
		notifyStatusActivity();
    }
    
    public void enableWifi(netControlState nextState) {
    	boolean done=false;
    	while (!done && !Thread.currentThread().isInterrupted()) {
			this.wifiManager.setWifiEnabled(true);
			// Waiting for interface-shutdown
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				// nothing
			}
			if (this.wifiManager.isWifiEnabled())
				done=true;
			else
				done=false;
    	}

		Log.d(MSG_TAG, "Wifi started!");
		curState=nextState;
		notifyStatusActivity();
    }
    
	public void enableAdhocServer(netControlState nextState) {
		// Ad-hoc mode, act as a DHCP Server
        this.coretask.updateDnsmasqConf();
		lastSwitchTime=System.currentTimeMillis();
        // Starting service
    	if (this.coretask.runRootCommand(this.coretask.DATA_FILE_PATH+"/bin/netcontrol start_server "+this.coretask.DATA_FILE_PATH)) {
    		Log.d(MSG_TAG, "netcontrol start_server failed");
    		// fall down below anyway
    	}
    	Log.d(MSG_TAG, "Adhoc DHCP Server enabled");
		curState=nextState;
		notifyStatusActivity();
	}

	public void disableAdhocServer(netControlState nextState) {
    	if (this.coretask.runRootCommand(this.coretask.DATA_FILE_PATH+"/bin/netcontrol stop_server "+this.coretask.DATA_FILE_PATH)) {
    		Log.d(MSG_TAG, "netcontrol stop_server failed");
    	}
		Log.d(MSG_TAG, "Adhoc DHCP Server disabled");
		curState=nextState;
		notifyStatusActivity();
	}

	public void enableAdhocClient(netControlState nextState) {
		// Ad-hoc mode, act as a DHCP Client
		Log.d(MSG_TAG, "Adhoc DHCP Client enabled");
		curState=nextState;
		notifyStatusActivity();
		discovered=false;
		myDhcpAddress="";
		lastSwitchTime=System.currentTimeMillis();
		if (this.coretask.runRootCommand(this.coretask.DATA_FILE_PATH+"/bin/netcontrol start_client "+this.coretask.DATA_FILE_PATH)) {
    		Log.d(MSG_TAG, "netcontrol start_client failed");
    		// doesn't necessarily mean the lease failed...
    	}

		myDhcpAddress=this.coretask.getMyDhcpAddress();
		if (myDhcpAddress.length() > 0) {
			discovered=true;
		}
	}

	public void disableAdhocClient(netControlState nextState) {
    	if (this.coretask.runRootCommand(this.coretask.DATA_FILE_PATH+"/bin/netcontrol stop_client "+this.coretask.DATA_FILE_PATH)) {
    		Log.d(MSG_TAG, "netcontrol stop_client failed");
    	}
    	discovered=false;
		Log.d(MSG_TAG, "Adhoc DHCP Client disabled");
		curState=nextState;
		notifyStatusActivity();
	}
	
	public void startService() {
		serviceStarted=true;
		switch (curState) {
    	case NOT_RUNNING:
    		if (getAdhocMode()) {
    			curState=netControlState.ADHOC_CLIENT;
    		}
    		else {
    			// Should already be enabled, but make sure
    			curState=netControlState.ACCESS_POINT;
    		}
    		notifyStatusActivity();
    		break;

    	case ADHOC_SERVER:
    	case ADHOC_CLIENT:
    	case ACCESS_POINT:
    		// Unexpected state
    		break;
		}
	}
	
	public void stopService() {
		serviceStarted=false;
	}
	
	public void restartService() {
		switch (curState) {
    	case NOT_RUNNING:
    		break;
    	case ADHOC_SERVER:
    		// Setting changed... shutdown adhoc mode
    		disableAdhocServer(curState);
			enableAdhocServer(curState);
			break;
    	case ADHOC_CLIENT:
    		disableAdhocClient(curState);
			enableAdhocClient(curState);
			break;
    	case ACCESS_POINT:
    		break;
		}		
	}
	
    class NetworkControl implements Runnable {
        // @Override
        public void run() {
        	boolean myServiceStarted=serviceStarted;
        	boolean stateChange=false;
        	Looper.prepare();
        	while (!Thread.currentThread().isInterrupted()) {
        		boolean newServiceState=serviceStarted;
        		if (myServiceStarted != newServiceState)
        			stateChange=true;
        		
	        	switch (curState) {
	        	case NOT_RUNNING:
	        		if (stateChange && newServiceState) {
		        		if (getAdhocMode()) {
		        			disableWifi(curState);
		        			enableAdhocClient(netControlState.ADHOC_CLIENT);
		        		}
		        		else {
		        			// Should already be enabled, but make sure
		        			enableWifi(netControlState.ACCESS_POINT);
		        		}
	        		}
	        		break;

	        	case ADHOC_SERVER:
	        		Log.d(MSG_TAG, "Adhoc Server: " +
	        				"friendList.size=" + Integer.toString(friendList.size()) +
	        				"staleFriendCout=" + Integer.toString(staleFriendCount) +
	        				"time delta=" + Long.toString(System.currentTimeMillis() - lastSwitchTime));

	        		if (stateChange) {
	        			Log.d(MSG_TAG, "Server: State change. newServiceState=" +
	        					(newServiceState ? "true" : "false"));
	        			if (newServiceState) {
	        				disableWifi(curState);
	        				enableAdhocServer(curState);
	        			}
	        			else {
	        				disableAdhocClient(curState);
	        				enableWifi(netControlState.NOT_RUNNING);	        				
	        			}
	        		}
	        		else if (!getAdhocMode()) {
	        			// Setting changed... shutdown adhoc mode
	        			disableAdhocServer(curState);
	        			enableWifi(netControlState.ACCESS_POINT);
	        		}
	        		else if ((friendList.size() == staleFriendCount) &&
	        				(System.currentTimeMillis() - lastSwitchTime) > SERVER_TIME) {
	        			disableAdhocServer(curState);
	        			enableAdhocClient(netControlState.ADHOC_CLIENT);
	        		}
	        		break;

	        	case ADHOC_CLIENT:
	        		if (stateChange) {
	        			Log.d(MSG_TAG, "Client: State change. newServiceState=" +
	        					(newServiceState ? "true" : "false"));
	        			if (newServiceState) {
	        				disableWifi(curState);
	        				enableAdhocClient(curState);
	        			} else {
	        				disableAdhocClient(curState);
	        				enableWifi(netControlState.NOT_RUNNING);
	        			}
	        		}
	        		else if (!getAdhocMode()) {
	        			// Setting changed... shutdown adhoc mode
	        			disableAdhocClient(curState);
	        			enableWifi(netControlState.ACCESS_POINT);
	        		}
	        		else {
		        		Log.d(MSG_TAG, "Adhoc Client discovered=" +
		        				(discovered ? "true" : "false") +
		        				"friendList.size=" + Integer.toString(friendList.size()) +
		        				"staleFriendCout=" + Integer.toString(staleFriendCount) +
		        				"time delta=" + Long.toString(System.currentTimeMillis() - lastSwitchTime));
		        		if (!discovered || (friendList.size() == staleFriendCount) &&
		        				(System.currentTimeMillis() - lastSwitchTime) > CLIENT_TIME) {
		        			// There are no active friends in the friend list
		            		disableAdhocClient(curState);
		            		enableAdhocServer(netControlState.ADHOC_SERVER);
		        		}
	        		}
	        		break;

	        	case ACCESS_POINT:
	        		if (stateChange) {
	        			Log.d(MSG_TAG, "Access Point: State change. newServiceState=" +
	        					(newServiceState ? "true" : "false"));
	        			if (newServiceState) {
		        			// Should already be enabled, but make sure
		        			enableWifi(curState);
	        			}
	        			else {
		        			// Should already be enabled, but make sure
		        			enableWifi(netControlState.NOT_RUNNING);
	        			}
	        		}
	        		else if (getAdhocMode()) {
	        			// Setting changed... shutdown wifi interface,
	        			// start adhoc mode
	        			disableWifi(curState);
	        			enableAdhocClient(netControlState.ADHOC_CLIENT);
	        		}
	        		break;
	        	}
	        	
	        	myServiceStarted=newServiceState;
	        	stateChange=false;
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
        	}
        }
    }
	
    class ClientSender implements Runnable {

        // @Override
        public void run() {
            byte[] buf=new byte[1024];
            boolean doSend=false;
        	DatagramPacket pkt=null;
       		DatagramSocket sock=null;
    		InetAddress dest=null;
    	    
        	Looper.prepare();
        	while (!Thread.currentThread().isInterrupted()) {
                try {
                	doSend=false;
                	if (curState == netControlState.NOT_RUNNING) {
                	}
                	if (curState == netControlState.ADHOC_CLIENT && discovered) {
                		Log.d(MSG_TAG, "Adhoc Client, discovered.");
                		Log.d(MSG_TAG, "myDhcpAddress=" + myDhcpAddress);
                		setMyIPAddress(myDhcpAddress);
                		int pos=myDhcpAddress.lastIndexOf('.');
                		String myBcastAddress=myDhcpAddress.substring(0, pos) + ".255";
                		Log.d(MSG_TAG, "myBcastAddress=" + myDhcpAddress);
                		dest=InetAddress.getByName(myBcastAddress);
                		doSend=true;
                	}
                	if (curState == netControlState.ADHOC_SERVER) {
                		Log.d(MSG_TAG, "Adhoc Server");
                		setMyIPAddress("192.168.2.1");
                		dest=InetAddress.getByName("192.168.2.255");
                		doSend=true;
                	}
                	if (curState == netControlState.ACCESS_POINT) {
                		Log.d(MSG_TAG, "Access Point.");
                		WifiInfo wifiInfo = wifiManager.getConnectionInfo();
                		int ipAddress = wifiInfo.getIpAddress();
                		String strAddress = toInetAddress(ipAddress);
                		setMyIPAddress(strAddress);
                		Log.d(MSG_TAG, "IP address="+strAddress);
                		// Make a broadcast address by masking the last 8 bits
                		ipAddress |= 0xff000000;
                		strAddress = toInetAddress(ipAddress);
                		dest=InetAddress.getByName(strAddress);
                		Log.d(MSG_TAG, "dest address="+strAddress);
                		doSend=true;
                	}
                		
                	if (doSend) {
                	    try {
                    		String msg=myData.makeHeartbeat(getMyGPSString());
                    	    buf = msg.getBytes();
                			pkt=new DatagramPacket(buf, buf.length, dest, 8888);
                	    	sock=new DatagramSocket();
                	   		sock.setBroadcast(true);
    	            		sock.send(pkt);
    	            		Log.d(MSG_TAG, "Packet was sent.");
                    	}
                    	catch (Exception e) {
                			// nothing
                    		Log.d(MSG_TAG, "Couldn't setup socket: " + e.getMessage());
                		}

                	}
            	}
            	catch (Exception e) {
        			// nothing
            		Log.d(MSG_TAG, "Couldn't send 192.168.2.255: " + e.toString());
            		e.printStackTrace();
        		}
            	
            	checkStaleFriends();
            	
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
        	}
        }
    }
    
    class ClientReceiver implements Runnable {

        // @Override
        public void run() {
            byte[] buf=new byte[1024];
        	DatagramPacket pkt=null;
       		DatagramSocket sock=null;
       		
        	Looper.prepare();
        	
        	boolean done=false;
        	while (!done && !Thread.currentThread().isInterrupted()) {
        		done=true;
        		try {
        			sock = new DatagramSocket(8888); 
            	}
    			catch (Exception e) {
    				e.printStackTrace();
    				done=false;
    			}
    			if (!done) {
                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        Thread.currentThread().interrupt();
                    }
    			}
        	}
    	        		
        	while (!Thread.currentThread().isInterrupted()) {
                try {
                	pkt = new DatagramPacket(buf, buf.length);
                	sock.receive(pkt);
                
                    String rcv = new String(pkt.getData(), 0, pkt.getLength());                    

            		Log.d(MSG_TAG, "Packet was received "  +
              			  pkt.getAddress().toString() + ": " +
            			  rcv );
            		
            		if (myData.IPaddress==null) {
            			// WTF
            		}
            		else if (pkt.getAddress().getHostAddress().compareTo(myData.IPaddress) == 0) {
                   		Log.d(MSG_TAG, "Packet was received from myself");       
            		}
            		else {
            			FriendData f=new FriendData();
               			f.IPaddress=pkt.getAddress().getHostAddress();
            			if (f.isHeartbeat(rcv)) {
            				String gps=f.parseHeartbeat(rcv);
            				processHeartbeat(f, gps);
            			}
            			Log.d(MSG_TAG, "isHeartbeat returned: " + (f.isHeartbeat(rcv) ? "true" : "false"));
            			if (f.isMessage(rcv)) {
            				String m=f.parseMessage(rcv);
            				processMessage(f, m);
            			}
            			Log.d(MSG_TAG, "isMessage returned: " + (f.isMessage(rcv) ? "true" : "false"));
            		}
                }
            	catch (IOException e) {
        			// nothing
            		Log.d(MSG_TAG, "Couldn't receive : " + e.getMessage());
        		}
        	}
        }
    }

    public synchronized void processHeartbeat(FriendData f, String gps) {

    	boolean found=false;
		Collection<FriendData> c = activeFriends.values();
		Iterator<FriendData> itr = c.iterator();
		int i=0;
		while(itr.hasNext()) {
			FriendData curFriend=itr.next();
			if (curFriend.equals(f)) {
	    		Log.d(MSG_TAG, "Updating existing friend: " + f.name + ", " + f.IPaddress);
	    		if (curFriend.lastHeartbeat == 0) {
	    			staleFriendCount--;
	    			notifyStatusActivity();
	    		}
				curFriend.lastHeartbeat=System.currentTimeMillis();
				if (curFriend.statusMsg.compareTo(f.statusMsg) != 0) {
					addEvent(String.valueOf(f.name + " status is now: " + f.statusMsg));
				}
				curFriend.statusMsg=f.statusMsg;
				friendList.set(i, String.valueOf(f.name + ": " + f.statusMsg + "\n" + "GPS: " + gps));
	    		notifyFriendListActivity();
	    		found=true;
				break;
			}
			i++;
		}
    	if (!found) {
    		Log.d(MSG_TAG, "Added a new friend: " + f.name + ", " + f.IPaddress);
    		// Found a new friend
    		int new_index=friendList.size();
    		f.lastHeartbeat=System.currentTimeMillis();
    		addEvent(String.valueOf(f.name + " status is now: " + f.statusMsg));
    		friendList.add(String.valueOf(f.name + ": " + f.statusMsg + "\n" + "GPS: " + gps));
    		activeFriends.put(new_index, f);
    		notifyFriendListActivity();
    		notifyStatusActivity();
    	}
    }
    
    public synchronized void checkStaleFriends() {
		Collection<FriendData> c = activeFriends.values();
		Iterator<FriendData> itr = c.iterator();
		int i=0;
		while(itr.hasNext()) {
			FriendData curFriend=itr.next();
			if (curFriend.lastHeartbeat > 0 &&
			    (System.currentTimeMillis()-curFriend.lastHeartbeat) > 15000)  {
	    		Log.d(MSG_TAG, "Marking stale friend: " + curFriend.name + ", " + curFriend.IPaddress);
				curFriend.lastHeartbeat=0;
				curFriend.statusMsg="unknown";
				staleFriendCount++;
				friendList.set(i, String.valueOf(curFriend.name + ": " + curFriend.statusMsg + " [stale]"));
	    		notifyFriendListActivity();
	    		notifyStatusActivity();
			}
			i++;
		}
    }
    
    public synchronized void processMessage(FriendData f, String message) {
    	addEvent(String.valueOf("Message from " + f.name + " at " + f.IPaddress));
    	messageHeader.add(eventList.get(0));
    	messageList.add(message);
    }
    
    public synchronized int findMessage(String event) {   

		ListIterator<String> itr = messageHeader.listIterator();
		int i=0;
		while(itr.hasNext()) {
			String curMessage=itr.next();
			if (event.compareTo(curMessage) == 0) {
				// found the matching event
				return i;
			}
			i++;
	    }
		
		return -1;
    }
    
    public synchronized void sendMessage(FriendData f, String message) {
    	String msgData=f.makeMessage(message);
        byte[] buf=new byte[msgData.length()];
    	DatagramPacket pkt=null;
   		DatagramSocket sock=null;
		InetAddress dest=null;
		try {
			dest=InetAddress.getByName(f.IPaddress);
    	    buf = msgData.getBytes();
			pkt=new DatagramPacket(buf, buf.length, dest, 8888);
	    	sock=new DatagramSocket();
    		sock.send(pkt);
    		Log.d(MSG_TAG, "Packet was sent.");
		}
		catch (Exception e) {
    		Log.d(MSG_TAG, "Couldn't send message: " + e.toString());
    		e.printStackTrace();
		}
    }
    
    public synchronized String getMessageHeader(int i) {
    	return messageHeader.get(i);
    }
    public synchronized String getMessage(int i) {
    	return messageList.get(i);
    }
    
	public synchronized List<String> getEventList() {
		return eventList;
	}

	public synchronized List<String> getFriendList() {
		return friendList;
	}

	public synchronized List<String> getMessageList() {
		return messageList;
	}

	public synchronized HashMap<Integer, FriendData> getActiveFriends() {
		return activeFriends;
	}
	
	public synchronized netControlState getCurrentState() {
		 return curState;
	}
	
	public synchronized int numActiveFriends() {
		return (friendList.size() - staleFriendCount);		
	}
	
	
	public String getMyGPSString() {
		String myLoc=Double.toString(dummyLat++) + ", " + Double.toString(dummyLong++);
		
		try {
			Location location=locationManager.getLastKnownLocation(LocationManager.GPS_PROVIDER);
			Double lat = location.getLatitude();
			Double lng = location.getLongitude();
			myLoc=Double.toString(lat) + ", " + Double.toString(lng);
		}
		catch (Exception e) {
			// GPS not available... use default string
		}
		return myLoc;
	}
	private static String toInetAddress(int ipAddress)
    {
        long ip = (ipAddress < 0) ? (long)Math.pow(2,32)+ipAddress : ipAddress;
        String addr =  String.valueOf((ip & 0xff)+"."+((ip >> 8) & 0xff)+"."+((ip >> 16) & 0xff)+"."+(ip >> 24));
        return addr;
    }
}